﻿/*	Application		IPC sender-side using socket (UDP)		
	File			Program.cs (IPC_socket_Sender)
	Purpose			Demonstration of local IPC, using a blocking socket (on the receiving side)
	Author			Richard John Anthony
	Date			January 2015
	
	Function		1. Creates a socket
					2. Waits for user to type a message
					3. Sends the message to the the application port (8007) on the local computer (uses local loopback address 127.0.0.1)
*/

using System;
using System.Text;
using System.Net;
using System.Net.Sockets;

namespace IPC_socket_Sender_with_Reply
{
    class C_IPC_socket_Sender
    {
        public const Int32 SEND_PORT = 8007;
        Socket m_IPC_SendSOCKET;
        public byte[] m_szSendBuf;
        IPEndPoint m_SendEndPoint;      // The address and port to send to

        static void Main(string[] args)
        {
            Console.WriteLine("Local Inter-Process Communication (sender side)");
            Console.WriteLine("This program waits for the user to type a message,");
            Console.WriteLine("and sends it to port 8007 on the local computer.");
            Console.WriteLine("------------------------------------------------------");
            Console.WriteLine("Type a  message to send ...   ");

            C_IPC_socket_Sender IPC_socket_Sender = new C_IPC_socket_Sender();
        }

        C_IPC_socket_Sender()
        {
            bool bSuccess;
            string sMessageString = Console.ReadLine(); // Get a string which may include spaces

            // Create Send socket
            bSuccess = CreateSendSocket();
            if (false == bSuccess)
            {
                Console.WriteLine("Exiting the program because a socket could not be created");
                CloseSocketAndExit();
            }

            // Set the Send Address details (Use local loopback IP address (127.0.0.1), and port 8007)
            SetUpSendAddressStructFor_LocalLoopback();

            // Send the message that the user typed, to the sepcified port on the local computer
            SendMessage(sMessageString);
            Console.WriteLine("   Message sent");

            CloseSocketAndExit();
        }

        bool CreateSendSocket()
        {
            try
            {
                m_IPC_SendSOCKET = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
            }
            catch (SocketException)  // Catch exceptions from the socket operations
            {
                Console.WriteLine("CreateSocket() failed");
                return false;
            }
            return true;
        }

        void SetUpSendAddressStructFor_LocalLoopback()
        {
            m_SendEndPoint = new System.Net.IPEndPoint(IPAddress.Loopback, System.Convert.ToInt16(SEND_PORT));
        }

        void SendMessage(string sSendString)
        {
            // Convert sSendString into the reply byte array format (!C# chars are two bytes each!)
            m_szSendBuf = Encoding.ASCII.GetBytes(sSendString);

            try
            {

                m_IPC_SendSOCKET.SendTo(m_szSendBuf, m_SendEndPoint);
            }
            catch (SocketException ex)
            {
                Console.WriteLine("sendto() Failed!");
                CloseSocketAndExit();
            }
        }

        void CloseSocketAndExit()
        {
            m_IPC_SendSOCKET.Close();
            Console.WriteLine("Socket closed");
            Console.WriteLine("(press a key to continue)");
            Console.ReadKey();	// Wait for user to type a key (to allow them to read the message before the window is closed)
            System.Environment.Exit(1);
        }
    }
}