﻿/*	Application		Periodic Output
	File			Program.cs (PeriodicOutput.cpp) 
	Purpose			Demonstration of command-line configurable program, with periodic output
	Author			Richard John Anthony
	Date			January 2015
	
	Function	1. Reads command-line parameters which configure the program's behaviour
				2. Outputs a user-specified character
				3. Sleeps a user-defined period
				4. Repeats in loop for as many iterations as necessary until the total run time reaches 10 seconds
*/

using System;

namespace PeriodicOutput
{
    class Program
    {
        static void Main(string[] args)
        {
            int iInterval;
            int iTotalRunTime_Milliseconds = 10000;		// Program will run for 10 seconds
            int iAccumulatedRunTime_Milliseconds = 0;	// Run time accumulated
            char cCharToDisplay;

            if (2 != args.Length)
            {
                // User did not provide the required parameters
                // Note, in C++ argv[0] is the program name, argv[1] is the first supplied parameter
                // However in C# args[0] is the first supplied parameter; so need a different technique to get the program name
                string sProgramName = Environment.GetCommandLineArgs()[0];  // Not the same as args[0]
                Console.WriteLine("Usage: {0}  milliseconds_period character_to_print", sProgramName);
                return;
            }

            iInterval = Convert.ToInt32(args[0]);
            if (0 >= iInterval)
            {
                // iInterval too small (less than or equal to 0)
                Console.WriteLine("Interval must be greater or equal to 1 millisecond");
                return;
            }

            cCharToDisplay = args[1][0];	// Get the first of possibly many chars that constitute the second argument

            while (iAccumulatedRunTime_Milliseconds < iTotalRunTime_Milliseconds)
            {
                Console.Write(cCharToDisplay);
                System.Threading.Thread.Sleep(iInterval);
                iAccumulatedRunTime_Milliseconds += iInterval;
            }
        }
    }
}

