﻿/*	Application		Threads Demonstration - Threads executed synchronously		
	File			program.cs (Threads_Join)
	Purpose			Demonstrates the basic concepts of multi-threading
	Author			Richard John Anthony
	Date			January 2015
	
	Function		1. Creates three threads
					2. Runs the threads synchronously - using Join() - the calling thread waits for the called thread to terminate
					3. Threads display text messages for diagnostic purposes
*/

using System;
using System.Threading;

class CThreads_Join
{ 
    public static void ThreadFunction1(object config)
    {
        try
        {
            ParameterizedThreadStart pts2 = new ParameterizedThreadStart(ThreadFunction2);
            Thread thread2 = new Thread(pts2);
            thread2.Start("BBBBBBBBBBBB");
            thread2.Join(); // The calling thread (thread1) waits for the called thread (thread2) to terminate          
            int iThreadID = Thread.CurrentThread.ManagedThreadId;
            string sStr = string.Format("F1 Thread ID:{0} data:{1}", iThreadID, (string)config);
            for (int iCount = 0; iCount < 10; iCount++)
            {
                Console.WriteLine(sStr);
            }
        }
        catch (Exception ex)
        {
        }
    }

    public static void ThreadFunction2(object config)
    {
        try
        {
            ParameterizedThreadStart pts3 = new ParameterizedThreadStart(ThreadFunction3);
            Thread thread3 = new Thread(pts3);
            thread3.Start("CCCCCCCCCCCC");
            thread3.Join(); // The calling thread (thread2) waits for the called thread (thread3) to terminate
            int iThreadID = Thread.CurrentThread.ManagedThreadId;
            string sStr = string.Format("F2 Thread ID:{0} data:{1}", iThreadID, (string)config);
            for (int iCount = 0; iCount < 10; iCount++)
            {
                Console.WriteLine(sStr);
            }
        }
        catch (Exception ex)
        {
        }
    }

    public static void ThreadFunction3(object config)
    {
        try
        {
            int iThreadID = Thread.CurrentThread.ManagedThreadId;
            string sStr = string.Format("F3 Thread ID:{0} data:{1}", iThreadID, (string)config);
            for (int iCount = 0; iCount < 10; iCount++)
            {
                Console.WriteLine(sStr);
            }
        }
        catch (Exception ex)
        {
        }
    }

    static void Main()
    {
        // Create three threads
        ParameterizedThreadStart pts1 = new ParameterizedThreadStart(ThreadFunction1);
        Thread thread1 = new Thread(pts1);
        thread1.Start("AAAAAAAAAAAA");
        thread1.Join(); // The calling thread (main) waits for the called thread (thread1) to terminate   

        for (int iCount = 0; iCount < 10; iCount++)
        {
            Console.WriteLine("Main");
        }

        Console.WriteLine("To Exit press any key, followed by the Enter key ");
        Console.ReadKey(true); // Wait for user to press a key (so the output results can be examined);
    }
}
